
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "config.h"
#include "hdy-enums-private.h"
#include "hdy-stackable-box-private.h"
#include "hdy-settings-private.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "hdy-settings-private.h" */

GType
hdy_system_color_scheme_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (HDY_SYSTEM_COLOR_SCHEME_DEFAULT), "HDY_SYSTEM_COLOR_SCHEME_DEFAULT", "default" },
        { C_ENUM (HDY_SYSTEM_COLOR_SCHEME_PREFER_DARK), "HDY_SYSTEM_COLOR_SCHEME_PREFER_DARK", "prefer-dark" },
        { C_ENUM (HDY_SYSTEM_COLOR_SCHEME_PREFER_LIGHT), "HDY_SYSTEM_COLOR_SCHEME_PREFER_LIGHT", "prefer-light" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("HdySystemColorScheme"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "hdy-stackable-box-private.h" */

GType
hdy_stackable_box_transition_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (HDY_STACKABLE_BOX_TRANSITION_TYPE_OVER), "HDY_STACKABLE_BOX_TRANSITION_TYPE_OVER", "over" },
        { C_ENUM (HDY_STACKABLE_BOX_TRANSITION_TYPE_UNDER), "HDY_STACKABLE_BOX_TRANSITION_TYPE_UNDER", "under" },
        { C_ENUM (HDY_STACKABLE_BOX_TRANSITION_TYPE_SLIDE), "HDY_STACKABLE_BOX_TRANSITION_TYPE_SLIDE", "slide" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("HdyStackableBoxTransitionType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

