
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

            #pragma once

            #include <glib-object.h>
            #if !defined(_HANDY_INSIDE) && !defined(HANDY_COMPILATION)
#error "Only <handy.h> can be included directly."
#endif

#include "hdy-enums.h"

            G_BEGIN_DECLS

/* enumerations from "hdy-settings-private.h" */

GType hdy_system_color_scheme_get_type (void);
#define HDY_TYPE_SYSTEM_COLOR_SCHEME (hdy_system_color_scheme_get_type())

/* enumerations from "hdy-stackable-box-private.h" */

GType hdy_stackable_box_transition_type_get_type (void);
#define HDY_TYPE_STACKABLE_BOX_TRANSITION_TYPE (hdy_stackable_box_transition_type_get_type())

G_END_DECLS

/* Generated data ends here */

