
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "config.h"
#include "hdy-enums.h"
#include "hdy-deck.h"
#include "hdy-flap.h"
#include "hdy-header-bar.h"
#include "hdy-header-group.h"
#include "hdy-leaflet.h"
#include "hdy-navigation-direction.h"
#include "hdy-squeezer.h"
#include "hdy-style-manager.h"
#include "hdy-tab-bar.h"
#include "hdy-view-switcher.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "hdy-deck.h" */

GType
hdy_deck_transition_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (HDY_DECK_TRANSITION_TYPE_OVER), "HDY_DECK_TRANSITION_TYPE_OVER", "over" },
        { C_ENUM (HDY_DECK_TRANSITION_TYPE_UNDER), "HDY_DECK_TRANSITION_TYPE_UNDER", "under" },
        { C_ENUM (HDY_DECK_TRANSITION_TYPE_SLIDE), "HDY_DECK_TRANSITION_TYPE_SLIDE", "slide" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("HdyDeckTransitionType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "hdy-flap.h" */

GType
hdy_flap_fold_policy_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (HDY_FLAP_FOLD_POLICY_NEVER), "HDY_FLAP_FOLD_POLICY_NEVER", "never" },
        { C_ENUM (HDY_FLAP_FOLD_POLICY_ALWAYS), "HDY_FLAP_FOLD_POLICY_ALWAYS", "always" },
        { C_ENUM (HDY_FLAP_FOLD_POLICY_AUTO), "HDY_FLAP_FOLD_POLICY_AUTO", "auto" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("HdyFlapFoldPolicy"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
hdy_flap_transition_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (HDY_FLAP_TRANSITION_TYPE_OVER), "HDY_FLAP_TRANSITION_TYPE_OVER", "over" },
        { C_ENUM (HDY_FLAP_TRANSITION_TYPE_UNDER), "HDY_FLAP_TRANSITION_TYPE_UNDER", "under" },
        { C_ENUM (HDY_FLAP_TRANSITION_TYPE_SLIDE), "HDY_FLAP_TRANSITION_TYPE_SLIDE", "slide" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("HdyFlapTransitionType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "hdy-header-bar.h" */

GType
hdy_centering_policy_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (HDY_CENTERING_POLICY_LOOSE), "HDY_CENTERING_POLICY_LOOSE", "loose" },
        { C_ENUM (HDY_CENTERING_POLICY_STRICT), "HDY_CENTERING_POLICY_STRICT", "strict" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("HdyCenteringPolicy"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "hdy-header-group.h" */

GType
hdy_header_group_child_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (HDY_HEADER_GROUP_CHILD_TYPE_HEADER_BAR), "HDY_HEADER_GROUP_CHILD_TYPE_HEADER_BAR", "header-bar" },
        { C_ENUM (HDY_HEADER_GROUP_CHILD_TYPE_GTK_HEADER_BAR), "HDY_HEADER_GROUP_CHILD_TYPE_GTK_HEADER_BAR", "gtk-header-bar" },
        { C_ENUM (HDY_HEADER_GROUP_CHILD_TYPE_HEADER_GROUP), "HDY_HEADER_GROUP_CHILD_TYPE_HEADER_GROUP", "header-group" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("HdyHeaderGroupChildType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "hdy-leaflet.h" */

GType
hdy_leaflet_transition_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (HDY_LEAFLET_TRANSITION_TYPE_OVER), "HDY_LEAFLET_TRANSITION_TYPE_OVER", "over" },
        { C_ENUM (HDY_LEAFLET_TRANSITION_TYPE_UNDER), "HDY_LEAFLET_TRANSITION_TYPE_UNDER", "under" },
        { C_ENUM (HDY_LEAFLET_TRANSITION_TYPE_SLIDE), "HDY_LEAFLET_TRANSITION_TYPE_SLIDE", "slide" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("HdyLeafletTransitionType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "hdy-navigation-direction.h" */

GType
hdy_navigation_direction_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (HDY_NAVIGATION_DIRECTION_BACK), "HDY_NAVIGATION_DIRECTION_BACK", "back" },
        { C_ENUM (HDY_NAVIGATION_DIRECTION_FORWARD), "HDY_NAVIGATION_DIRECTION_FORWARD", "forward" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("HdyNavigationDirection"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "hdy-squeezer.h" */

GType
hdy_squeezer_transition_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (HDY_SQUEEZER_TRANSITION_TYPE_NONE), "HDY_SQUEEZER_TRANSITION_TYPE_NONE", "none" },
        { C_ENUM (HDY_SQUEEZER_TRANSITION_TYPE_CROSSFADE), "HDY_SQUEEZER_TRANSITION_TYPE_CROSSFADE", "crossfade" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("HdySqueezerTransitionType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "hdy-style-manager.h" */

GType
hdy_color_scheme_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (HDY_COLOR_SCHEME_DEFAULT), "HDY_COLOR_SCHEME_DEFAULT", "default" },
        { C_ENUM (HDY_COLOR_SCHEME_FORCE_LIGHT), "HDY_COLOR_SCHEME_FORCE_LIGHT", "force-light" },
        { C_ENUM (HDY_COLOR_SCHEME_PREFER_LIGHT), "HDY_COLOR_SCHEME_PREFER_LIGHT", "prefer-light" },
        { C_ENUM (HDY_COLOR_SCHEME_PREFER_DARK), "HDY_COLOR_SCHEME_PREFER_DARK", "prefer-dark" },
        { C_ENUM (HDY_COLOR_SCHEME_FORCE_DARK), "HDY_COLOR_SCHEME_FORCE_DARK", "force-dark" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("HdyColorScheme"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "hdy-view-switcher.h" */

GType
hdy_view_switcher_policy_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (HDY_VIEW_SWITCHER_POLICY_AUTO), "HDY_VIEW_SWITCHER_POLICY_AUTO", "auto" },
        { C_ENUM (HDY_VIEW_SWITCHER_POLICY_NARROW), "HDY_VIEW_SWITCHER_POLICY_NARROW", "narrow" },
        { C_ENUM (HDY_VIEW_SWITCHER_POLICY_WIDE), "HDY_VIEW_SWITCHER_POLICY_WIDE", "wide" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("HdyViewSwitcherPolicy"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

