#pragma once

#include <handy.h>

G_BEGIN_DECLS

#define HDY_TYPE_TAB_VIEW_DEMO_WINDOW (hdy_tab_view_demo_window_get_type())

G_DECLARE_FINAL_TYPE (HdyTabViewDemoWindow, hdy_tab_view_demo_window, HDY, TAB_VIEW_DEMO_WINDOW, HdyWindow)

HdyTabViewDemoWindow *hdy_tab_view_demo_window_new (void);

void hdy_tab_view_demo_window_prepopulate (HdyTabViewDemoWindow *self);

G_END_DECLS
