/* This file was generated by gen-plublic-types.sh, do not edit it. */

#include "../libhandy/src/hdy-action-row.h"
#include "../libhandy/src/hdy-animation.h"
#include "../libhandy/src/hdy-application-window.h"
#include "../libhandy/src/hdy-avatar.h"
#include "../libhandy/src/hdy-carousel.h"
#include "../libhandy/src/hdy-carousel-indicator-dots.h"
#include "../libhandy/src/hdy-carousel-indicator-lines.h"
#include "../libhandy/src/hdy-clamp.h"
#include "../libhandy/src/hdy-combo-row.h"
#include "../libhandy/src/hdy-deck.h"
#include "../libhandy/src/hdy-enum-value-object.h"
#include "../libhandy/src/hdy-expander-row.h"
#include "../libhandy/src/hdy-flap.h"
#include "../libhandy/src/hdy-header-bar.h"
#include "../libhandy/src/hdy-header-group.h"
#include "../libhandy/src/hdy-keypad.h"
#include "../libhandy/src/hdy-leaflet.h"
#include "../libhandy/src/hdy-main.h"
#include "../libhandy/src/hdy-navigation-direction.h"
#include "../libhandy/src/hdy-preferences-group.h"
#include "../libhandy/src/hdy-preferences-page.h"
#include "../libhandy/src/hdy-preferences-row.h"
#include "../libhandy/src/hdy-preferences-window.h"
#include "../libhandy/src/hdy-search-bar.h"
#include "../libhandy/src/hdy-squeezer.h"
#include "../libhandy/src/hdy-status-page.h"
#include "../libhandy/src/hdy-style-manager.h"
#include "../libhandy/src/hdy-swipe-group.h"
#include "../libhandy/src/hdy-swipe-tracker.h"
#include "../libhandy/src/hdy-swipeable.h"
#include "../libhandy/src/hdy-tab-bar.h"
#include "../libhandy/src/hdy-tab-view.h"
#include "../libhandy/src/hdy-title-bar.h"
#include "../libhandy/src/hdy-types.h"
#include "../libhandy/src/hdy-value-object.h"
#include "../libhandy/src/hdy-view-switcher.h"
#include "../libhandy/src/hdy-view-switcher-bar.h"
#include "../libhandy/src/hdy-view-switcher-title.h"
#include "../libhandy/src/hdy-window.h"
#include "../libhandy/src/hdy-window-handle.h"
#include "src/hdy-enums.h"
#include "hdy-main-private.h"

void
hdy_init_public_types (void)
{
  g_type_ensure (HDY_TYPE_ACTION_ROW);
  g_type_ensure (HDY_TYPE_APPLICATION_WINDOW);
  g_type_ensure (HDY_TYPE_AVATAR);
  g_type_ensure (HDY_TYPE_CAROUSEL);
  g_type_ensure (HDY_TYPE_CAROUSEL_INDICATOR_DOTS);
  g_type_ensure (HDY_TYPE_CAROUSEL_INDICATOR_LINES);
  g_type_ensure (HDY_TYPE_CENTERING_POLICY);
  g_type_ensure (HDY_TYPE_CLAMP);
  g_type_ensure (HDY_TYPE_COLOR_SCHEME);
  g_type_ensure (HDY_TYPE_COMBO_ROW);
  g_type_ensure (HDY_TYPE_DECK);
  g_type_ensure (HDY_TYPE_DECK_TRANSITION_TYPE);
  g_type_ensure (HDY_TYPE_ENUM_VALUE_OBJECT);
  g_type_ensure (HDY_TYPE_EXPANDER_ROW);
  g_type_ensure (HDY_TYPE_FLAP);
  g_type_ensure (HDY_TYPE_FLAP_FOLD_POLICY);
  g_type_ensure (HDY_TYPE_FLAP_TRANSITION_TYPE);
  g_type_ensure (HDY_TYPE_HEADER_BAR);
  g_type_ensure (HDY_TYPE_HEADER_GROUP);
  g_type_ensure (HDY_TYPE_HEADER_GROUP_CHILD);
  g_type_ensure (HDY_TYPE_HEADER_GROUP_CHILD_TYPE);
  g_type_ensure (HDY_TYPE_KEYPAD);
  g_type_ensure (HDY_TYPE_LEAFLET);
  g_type_ensure (HDY_TYPE_LEAFLET_TRANSITION_TYPE);
  g_type_ensure (HDY_TYPE_NAVIGATION_DIRECTION);
  g_type_ensure (HDY_TYPE_PREFERENCES_GROUP);
  g_type_ensure (HDY_TYPE_PREFERENCES_PAGE);
  g_type_ensure (HDY_TYPE_PREFERENCES_ROW);
  g_type_ensure (HDY_TYPE_PREFERENCES_WINDOW);
  g_type_ensure (HDY_TYPE_SEARCH_BAR);
  g_type_ensure (HDY_TYPE_SQUEEZER);
  g_type_ensure (HDY_TYPE_SQUEEZER_TRANSITION_TYPE);
  g_type_ensure (HDY_TYPE_STATUS_PAGE);
  g_type_ensure (HDY_TYPE_STYLE_MANAGER);
  g_type_ensure (HDY_TYPE_SWIPEABLE);
  g_type_ensure (HDY_TYPE_SWIPE_GROUP);
  g_type_ensure (HDY_TYPE_SWIPE_TRACKER);
  g_type_ensure (HDY_TYPE_TAB_BAR);
  g_type_ensure (HDY_TYPE_TAB_PAGE);
  g_type_ensure (HDY_TYPE_TAB_VIEW);
  g_type_ensure (HDY_TYPE_TITLE_BAR);
  g_type_ensure (HDY_TYPE_VALUE_OBJECT);
  g_type_ensure (HDY_TYPE_VIEW_SWITCHER);
  g_type_ensure (HDY_TYPE_VIEW_SWITCHER_BAR);
  g_type_ensure (HDY_TYPE_VIEW_SWITCHER_POLICY);
  g_type_ensure (HDY_TYPE_VIEW_SWITCHER_TITLE);
  g_type_ensure (HDY_TYPE_WINDOW);
  g_type_ensure (HDY_TYPE_WINDOW_HANDLE);
}

